module code.common.Contexts where

open import Relation.Binary.PropositionalEquality

open import Data.Fin
open import Data.List
open import Data.Vec hiding (length ; head ; _++_)

-- There's a lot of kruft in here; I apologize.

--------------------
-- Contexts
--------------------

Cxt = List

-- Type context membership

infix 3 _∈_
data _∈_ {ℓ} {A : Set ℓ} (x : A) : List {ℓ} A → Set where
  zero : ∀ {xs} → x ∈ x ∷ xs
  suc : ∀ {y xs} → x ∈ xs → x ∈ y ∷ xs

-- Good, old-fashioned list lookup
lkup∈ : ∀ {A : Set} {x} → (ρ : List A) → (x ∈ ρ) → A
lkup∈ (x ∷ _) zero     = x
lkup∈ (_ ∷ xs) (suc i) = lkup∈ xs i

infixr 3 _▷_

data All {A : Set} (P : A → Set) : List A → Set where
  []  : All P []
  _▷_ : ∀ {x xs} → P x → All P xs → All P (x ∷ xs)

lookup∈ : ∀ {A} {P : A → Set} {x xs} → All P xs → x ∈ xs → P x
lookup∈ [] ()                            -- impossible case, but needed to make Agda happy
lookup∈ (px ▷ ps) zero   = px            -- actual environment verification
lookup∈ (_ ▷ ps) (suc i) = lookup∈ ps i

data _⊆_ {ℓ} { A : Set ℓ } : List {ℓ} A → List A → Set ℓ where
  ▪    : ∀ {η}
           -------
         → [] ⊆ η     -- no instance. slash-blacksquare
  inst : ∀ {x xs η}
         →  x ∈ η
         →  xs ⊆ η
           ------------
         → (x ∷ xs) ⊆ η

lkupμ : ∀ {A μ} → (xs : List A) → μ ⊆ xs → List A
lkupμ xs ▪          = []
lkupμ xs (inst x i) = (lkup∈ xs x) ∷ (lkupμ xs i)

desnorf : ∀ {A P x} → (xs : List A) → All P xs → (i : x ∈ xs) → P (lkup∈ xs i)
desnorf _ (px ▷ _) zero            = px
desnorf (_ ∷ xs) (_ ▷ pxs) (suc i) = desnorf xs pxs i

derefμ : ∀ {A μ P} → (xs : List A) → (i : μ ⊆ xs) → All P xs → All P (lkupμ xs i)
derefμ xs ▪ pxs          = []
derefμ xs (inst i μ) pxs = desnorf xs pxs i ▷ derefμ xs μ pxs

nudge : ∀ {ℓ} {A L x} → (l : List {ℓ} A) → (x ∈ L) → x ∈ (l ++ L)
nudge [] x      = x
nudge (_ ∷ l) x = suc (nudge l x)

nudge' : ∀ {ℓ} {A L x} → (l : List {ℓ} A) → (x ∈ L) → x ∈ (l ++ L)
nudge' [] x      = x
nudge' (_ ∷ l) x = suc (nudge' l x)

nlkup : ∀ {ℓ} {A : Set ℓ} {x} → (Γ : List A) → (x ∈ Γ) → A
nlkup (y ∷ ys) zero    = y
nlkup (y ∷ ys) (suc i) = nlkup ys i

nl : ∀ {ℓ} {A : Set ℓ} {x} {Γ : List A} → (Vec A (length Γ)) → (x ∈ Γ) → A
nl (y ∷ ys) zero    = y
nl (y ∷ ys) (suc i) = nl ys i

TheBigLemma : ∀ {ℓ} {A x} →
                (Γ : List {ℓ} A) → -- e.g., a Cxt
                (i : x ∈ Γ)      → -- rather than a Fin
                -----------------------------
                nlkup Γ i ≡ nl (fromList Γ) i
                         -- ^^^^^ lookup (compress Γ) identical
TheBigLemma .(_ ∷ _) zero                             = refl
TheBigLemma (_ ∷ xs) (suc i) rewrite TheBigLemma xs i = refl

∈2fin : ∀  {ℓ} {A : Set ℓ} {S : List {ℓ} A} {x} → x ∈ S → Fin (length S)
∈2fin zero    = zero
∈2fin (suc i) = suc (∈2fin i)

