module code.common.Idioms where

open import Agda.Primitive
open import Data.Product

-- some abbreviations and idioms
π₁ : ∀ {a b : Set} → (a × b) → a
π₁ = Data.Product.proj₁
π₂ : ∀ {a b : Set} → (a × b) → b
π₂ = Data.Product.proj₂

infix 3 _∘_

_∘_ : ∀ {a b c : Set} → (b → c) → (a → b) → a → c
f ∘ g = λ a → f (g a)

--
-- Abbreviations and local idioms
--

case_of_ : ∀ {a b} {A : Set a} {B : Set b} → A → (A → B) → B
case x of f = f x

π¹³ : ∀ {a b c : Set} → (a × b × c) → a
π¹³ abc = π₁ abc

π²³ : ∀ {a b c : Set} → (a × b × c) → b
π²³ abc = π₁ (π₂ abc)

π³³ : ∀ {a b c : Set} → (a × b × c) → c
π³³ abc = π₂ (π₂ abc)

π₃ : ∀ {a b c d : Set} → (a × b × c × d) → c
π₃ abcd = π₁ (π₂ (π₂ abcd)) 

π₄ : ∀ {a b c d : Set} → (a × b × c × d) → d
π₄ abcd = π₂ (π₂ (π₂ abcd)) 

π³⁴ : ∀ {a b c d : Set} → (a × b × c × d) → c
π³⁴ abcd = π₁ (π₂ (π₂ abcd)) 

π⁴⁴ : ∀ {a b c d : Set} → (a × b × c × d) → d
π⁴⁴ abcd = π₂ (π₂ (π₂ abcd)) 

data Nil : Set where -- doesn't seem to be built-in
  ⋆ : Nil

------------------------------------------------------------------------
-- A level polymorphic unit and empty type.

record ⊤ {ℓ : Level} : Set ℓ where
  constructor tt

record _≤_ {ℓ ℓ′ e : Level} (x : ⊤ {ℓ}) (y : ⊤ {ℓ′}) : Set e where
