module code.devicecalculus.MealyMachine where

open import Data.Product

open import code.common.Idioms 
open import code.common.MachineWords
open import code.common.Stream renaming (hd to shd; tl to stl) 

----------------------------------
----------------------------------
-- Mealy as a coinductive type  --
----------------------------------
----------------------------------

record Mealy (i : Set) (s : Set) (o : Set) : Set where
  coinductive
  field fun : i → s → (o × s)
        now : i × s × o
        nxt : i → Mealy i s o

open Mealy

unfoldM : ∀ {i s o} → (i → s → (o × s)) → i → s → o → Mealy i s o
fun (unfoldM f i s o)   = f
now (unfoldM f i s o)   = (i , s , o)
nxt (unfoldM f _ s o) i = let
                            (o' , s') = f i s 
                          in
                            unfoldM f i s' o'


----------------------------------------------
----------------------------------------------
-- Mealy Operators (not reflected in FIRRTL --
-- semantics (could be extensions).         --
----------------------------------------------
----------------------------------------------

fun× : ∀ {i1 i2 s1 s2 o1 o2 : Set} →
       (i1 → s1 → (o1 × s1)) →  
       (i2 → s2 → (o2 × s2)) →
       (i1 × i2) → (s1 × s2) → ((o1 × o2) × (s1 × s2))
fun× f1 f2 (i1 , i2) (s1 , s2) = let
                                    (o1' , s1') = f1 i1 s1
                                    (o2' , s2') = f2 i2 s2
                                 in
                                    ((o1' , o2') , (s1' , s2')) 

infix 4 _∥_

_∥_ : ∀ {i1 i2 s1 s2 o1 o2} →
        Mealy i1 s1 o1 →
        Mealy i2 s2 o2 →
        Mealy (i1 × i2) (s1 × s2) (o1 × o2)
m1 ∥ m2 = let
            (i1 , s1 , o1) = now m1
            (i2 , s2 , o2) = now m2
          in
             unfoldM (fun× (fun m1) (fun m2)) (i1 , i2) (s1 , s2) (o1 , o2)

wrap : ∀ {i1 i2 s o1 o2 : Set} →
      (o1 → o2)                →
      (i2 → o1 → i1)           →
      o1                       →
      (i1 → s → (o1 × s))      →
      (i2 → s → (o2 × s)) 
wrap out conn oi f ix s = let
                            ii         = conn ix oi
                            (oi' , s') = f ii s 
                          in
                            (out oi' , s')

feedbackM : ∀ {i1 i2 s o1 o2} →
          (o1 → o2)      →
          (i2 → o1 → i1) →
          i2             →
          Mealy i1 s o1  →
          Mealy i2 s o2
feedbackM out conn i2 m = let
                         (_ , s , o1) = now m
                         f2 = wrap out conn o1 (fun m) 
                        in
                         unfoldM f2 i2 s (out o1)

pipeline : ∀ {i1 s1 s2 o1 o2} →
           Mealy i1 s1 o1 →
           Mealy o1 s2 o2 →
           Mealy i1 (s1 × s2) o2
pipeline m1 m2 = feedbackM π₂ (λ i1 → λ (o1 , o2) → (i1 , o1)) (π¹³ (now m1)) (m1 ∥ m2)

open Stream
open import Data.Nat
open import Data.List
      
transcript : ∀ {i s o} → Mealy i s o → Stream i → Stream (i × s × o)
hd (transcript m is) = now m 
tl (transcript m is) = transcript (nxt m (hd is)) (tl is)
runM : ∀ {i s o} → ℕ → Mealy i s o → Stream i → List (i × s × o)
runM n m is = takeStr n (transcript m is)
