module code.devicecalculus.Types where

open import Data.Bool
open import Data.Fin renaming (zero to ZERO; suc to SUC) hiding (lift)
open import Data.Nat
open import Data.Vec renaming (lookup to vecref)
open import Data.Product
open import Data.Sum

open import code.common.Idioms 
open import code.common.MachineWords
open import code.devicecalculus.MealyMachine

--------------------
-- Type language
--------------------

data Type : Set where
  nat   : Type
  bool  : Type
  unit  : Type
  bit   : Type
  slv   : ℕ → Type -- standard logic vector a' la VHDL
  index : ℕ → Type
  _⇒_  : Type → Type → Type
  _⊗_  : Type → Type → Type
  _⊕_  : Type → Type → Type
  Dev  : Type → Type → Type → Type

---
--- Type Semantics
---

infixr 4 _⊗_
infixr 4 _⇒_
    
Value : Type → Set
Value nat            = ℕ
Value bool           = Bool
Value unit           = ⊤
Value bit            = Bool
Value (slv n)        = Vec Bit n
Value (index n)      = Fin n
Value (t1 ⇒ t2)      = (Value t1) → (Value t2)
Value (t1 ⊗ t2)      = (Value t1) × (Value t2)
Value (t1 ⊕ t2)      = (Value t1) ⊎ (Value t2)
Value (Dev i s o)     = Mealy (Value i) (Value s) (Value o)

--- 
--- End of Type Semantics
---
