module code.examples.Accumulator where

-- revisit the imports
open import Agda.Builtin.String renaming (primStringAppend to _+++_)
open import Agda.Primitive
open import Data.Bool renaming (Bool to Bit; false to 0#; true to 1#)
open import Data.Fin using (Fin; zero; suc)
open import Data.List hiding (lookup ; [_] ; and)
open import Data.Maybe hiding (map)
open import Data.Nat using (ℕ; _∸_; _<_ ; _≤_ ; _+_) renaming (zero to Z; suc to S)
open import Data.Product hiding (map ; swap ; curry ; uncurry)
open import Data.Vec hiding (foldl ; foldr ; _++_ ; map ; _[_]=_ ; length) renaming (lookup to vecref ; reverse to rev)

open import code.common.Idioms hiding (_∘_)
open import code.common.Contexts
open import code.common.MachineWords
open import code.common.Stream renaming (hd to shd; tl to stl ; map to mapStr ; zip to zipStr)

open import code.firrtl.Types
open import code.firrtl.Syntax
open import code.firrtl.TransCircuit

open import code.devicecalculus.Types
open import code.devicecalculus.Syntax
open import code.devicecalculus.MealyMachine
open import code.devicecalculus.Semantics

-- Some abbreviations

W1 : Ty
W1 = ↑ UInt 1

W8 : Ty
W8 = ↑ UInt 8


-- circuit Accumulator : @[:@2.0]
--   module Accumulator : @[:@3.2]
--     input clock : Clock @[:@4.4]
--     input reset : UInt<1> @[:@5.4]
--     input io_in : UInt<1> @[:@6.4]
--     output io_out : UInt<8> @[:@6.4]
--     reg accumulator : UInt<8>, clock with :
--       reset => (UInt<1>("h0"), accumulator) @[Accumulator.scala 16:28:@8.4]
--     node _T_11 = add(accumulator, io_in) @[Accumulator.scala 17:30:@9.4]
--     node _T_12 = tail(_T_11, 1) @[Accumulator.scala 17:30:@10.4]
--     io_out <= accumulator @[Accumulator.scala 18:10:@12.4]
--     accumulator <= mux(reset, UInt<8>("h0"), _T_12) @[Accumulator.scala 17:15:@11.4]
--     skip


-- |
-- | N.b., we don't allow automated coercions, so io_in is now a W8
-- |
Δ : ModTy
Δ = Mod (W1 ⊗ W8 ⊗ Unit) , (W8 ⊗ Unit) , (W8 ⊗ Unit)  ⋊ []

t11 : [] ⊢ₓ UInt 8 // Δ 
t11 = add (reg Zᵍ) (inp (Sᵍ Zᵍ))

newacc : [] ⊢ₓ UInt 8 // Δ
newacc = mux (inp Zᵍ) z8 t11
  where
    z8 : ∀ {Γ Δ} → Γ ⊢ₓ UInt 8 // Δ
    z8 = uint (0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ [])

acc_mod : [] ⊢ᵐ Δ
acc_mod = mod ◇
              (node none (⊗case (reg Zᵍ) unit))  -- outs
              (node none (⊗case newacc unit))    -- regs
              (node none unit)

accumulator : [] ⊢ᶜ Δ
accumulator = mod acc_mod

accumulatorTr : [] ⊩ᵈ Dev (is Δ) (rs Δ) (os Δ)
accumulatorTr = trC Δ [] accumulator

accumulator' : Mealy (Vec Bit 1 × Vec Bit 8 × ⊤) (Vec Bit 8 × ⊤) (Vec Bit 8 × ⊤)
accumulator' = ⟦ accumulatorTr ⟧ []

-- |
-- | Testing
-- |

-- |
-- | The application of reverse (rev) to these bitstrings is because the arithmetic operations
-- | defined in MachineWords.agda operate on words arranged from least- to most-significant bit.
-- | 

zero1 : Vec Bit 1
zero1 = (0# ∷ [])

one8 : Vec Bit 8
one8 = rev (0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 0# ∷ 1# ∷ [])

testinputs : Stream (Vec Bit 1 × Vec Bit 8 × ⊤ {lzero})
testinputs = cycle testhd testtl
   where
      testhd = (zero1 , one8 , tt)
      testtl = []

pair : String → String → String
pair i s = "(" +++ (i +++ (" , " +++ (s +++ ")")))

triple : String → String → String → String
triple i s o = "(" +++ (i +++ (" , " +++ (s +++ (" , " +++ (o +++ ")")))))

-- hideous
pp : (Vec Bit 1 × Vec Bit 8 × ⊤ {lzero}) × (Vec Bit 8 × ⊤ {lzero}) × (Vec Bit 8 × ⊤ {lzero}) → String
pp (( reset , ioin , _) , (acc , _) , (ioout , _)) = triple (pair (ppBV reset) (ppBV ioin)) (ppBV acc) (ppBV ioout)
    
-- | Matches the simulator exactly.
runaccumulator = Data.List.tail (map pp (takeStr 11 (transcript accumulator' testinputs)))

-- | Matches the simulator exactly.
{- 
("((0 , 00000001) , 00000001 , 00000000)" ∷
 "((0 , 00000001) , 00000010 , 00000001)" ∷
 "((0 , 00000001) , 00000011 , 00000010)" ∷
 "((0 , 00000001) , 00000100 , 00000011)" ∷
 "((0 , 00000001) , 00000101 , 00000100)" ∷
 "((0 , 00000001) , 00000110 , 00000101)" ∷
 "((0 , 00000001) , 00000111 , 00000110)" ∷
 "((0 , 00000001) , 00001000 , 00000111)" ∷
 "((0 , 00000001) , 00001001 , 00001000)" ∷
 "((0 , 00000001) , 00001010 , 00001001)" ∷ [])
-}
