module code.examples.Figure4 where

-- revisit the imports
open import Agda.Builtin.String renaming (primStringAppend to _+++_)
open import Agda.Primitive
open import Data.Bool renaming (Bool to Bit; false to 0#; true to 1#)
open import Data.Fin using (Fin; zero; suc)
open import Data.List hiding (lookup ; [_] ; and)
open import Data.Maybe hiding (map)
open import Data.Nat using (ℕ; _∸_; _<_ ; _≤_ ; _+_) renaming (zero to Z; suc to S)
open import Data.Product hiding (map ; swap ; curry ; uncurry)
open import Data.Vec hiding (foldl ; foldr ; _++_ ; map ; _[_]=_ ; length) renaming (lookup to vecref)

open import code.common.Idioms hiding (_∘_)
open import code.common.Contexts
open import code.common.MachineWords
open import code.common.Stream renaming (hd to shd; tl to stl ; map to mapStr ; zip to zipStr)

open import code.firrtl.Types
open import code.firrtl.Syntax
open import code.firrtl.TransCircuit

open import code.devicecalculus.MealyMachine
open import code.devicecalculus.Semantics

{-

Note to reviewers: this file contains the example from Figure 4.

fib4' near the bottom of the file is the relevant Mealy.

Evaluating "doit" will generate the output.

Δ : ModTy
Δ = Mod (W1 ⊗ W1) , (W4 ⊗ W4) , W4 ⋊ []

firfib4 : [] ⊢ᶜ Δ
circuit firfib4 :
  module firfib4 :
    input go    : UInt<1>
    input reset : UInt<1>
    input clk   : Clock
    reg n       : UInt<4>, clk
    reg m       : UInt<4>, clk
    output out  : UInt<4>

    n <= if reset then 0 
         elsif go then m else n
    m <= if reset then 1
         elsif go then n + m else m
    out <= n
-}

W4 : Ty
W4 = ↑ UInt 4

W1 : Ty
W1 = ↑ UInt 1

s4 : Ty
s4 = W4 ⊗ W4 ⊗ Unit

o4 : Ty
o4 = W4 ⊗ Unit

zero4 : ∀ {Γ Δ} → Γ ⊢ₓ UInt 4 // Δ
zero4 = uint (0# ∷ 0# ∷ 0# ∷ 0# ∷ [])

one4 : ∀ {Γ Δ} → Γ ⊢ₓ UInt 4 // Δ
one4 = uint (1# ∷ 0# ∷ 0# ∷ 0# ∷ [])

--
-- Defining resetfib4 in the FIRRTL abstract syntax.
-- 

Δf4 : ModTy
Δf4 = Mod (W1 ⊗ W1 ⊗ Unit) , (W4 ⊗ W4 ⊗ Unit) , W4 ⊗ Unit  ⋊ []

ofib4 : [] ⊢ₓ UInt 4 // Δf4
ofib4 = mux (inp (Sᵍ Zᵍ))
           zero4
           (reg Zᵍ)

regn4 : [] ⊢ₓ UInt 4 // Δf4
regn4 = mux (inp (Sᵍ Zᵍ))      -- if reset
            zero4              --   then n <- 0
            (mux (inp Zᵍ)      -- elsif go 
                 (reg (Sᵍ Zᵍ)) --    then n <- m
                 (reg Zᵍ))     --  else n <- n

regm4 : [] ⊢ₓ UInt 4 // Δf4
regm4 = mux (inp (Sᵍ Zᵍ))                   -- if reset 
           one4                             --   then m <- 1
           (mux (inp Zᵍ)                    -- elsif go
                (add (reg Zᵍ) (reg (Sᵍ Zᵍ))) --    then m <- n + m
                (reg (Sᵍ Zᵍ)))               --    else m <- m

firfib4 : [] ⊢ᶜ Δf4
firfib4 = mod (mod ◇ (node none (⊗case ofib4 unit))                 -- set outs
                       (node none (⊗case regn4 (⊗case regm4 unit))) -- set regs
                       (node none unit))

fib4' : Mealy (Vec Bit 1 × Vec Bit 1 × ⊤) (Vec Bit 4 × Vec Bit 4 × ⊤) (Vec Bit 4 × ⊤)
fib4' = ⟦ trC Δf4 [] firfib4 ⟧ []

seed0' : List {lzero} (Vec Bit 1 × Vec Bit 1 × ⊤)
seed0' = map (λ (b1 , b2) → (b1 ∷ [] , b2 ∷ [] , tt)) seed0

--
-- Hacking together pretty printing in Agda; there must be a better way.
--
pp' : ∀ {i j k l m} → ((Vec Bit i × Vec Bit j × ⊤ {lzero}) × (Vec Bit k × Vec Bit l × ⊤ {lzero}) × Vec Bit m × ⊤ {lzero}) → String
pp' ((b1 , b2 , _) , (w1 , w2 , _) , w3 , _) =
                                      "(" +++ (ppBV b1 +++ (" , " +++ (ppBV b2 +++ (") , (" +++
                                      (ppBV w1 +++ (" , " +++ (ppBV w2 +++ (") , " +++
                                      (ppBV w3 +++ ")")))))))))

ex4 : List ((Vec Bit 1 × Vec Bit 1 × ⊤) × (Vec Bit 4 × Vec Bit 4 × ⊤) × Vec Bit 4 × ⊤)
ex4 = runM 12 fib4' (cycle (0# ∷ [] , 1# ∷ [] , tt) seed0')

doit = map pp' ex4


{- doit running example output
"(0 , 0) , (0000 , 0000) , 0000)" ∷
"(0 , 1) , (0000 , 0001) , 0000)" ∷
"(1 , 0) , (0001 , 0001) , 0000)" ∷
"(0 , 0) , (0001 , 0001) , 0001)" ∷
"(1 , 0) , (0001 , 0010) , 0001)" ∷
"(0 , 0) , (0001 , 0010) , 0001)" ∷
"(1 , 0) , (0010 , 0011) , 0001)" ∷
"(0 , 0) , (0010 , 0011) , 0010)" ∷
"(1 , 0) , (0011 , 0101) , 0010)" ∷
"(0 , 0) , (0011 , 0101) , 0011)" ∷
"(1 , 0) , (0101 , 1000) , 0011)" ∷
"(0 , 1) , (0000 , 0001) , 0000)" ∷ []
-}
{- pretty printed by hand; doit running example output

((0 , 0) , (0000 , 0000) , 0000) ∷
((0 , 1) , (0000 , 0001) , 0000) ∷
((1 , 0) , (0001 , 0001) , 0000) ∷
((0 , 0) , (0001 , 0001) , 0001) ∷
((1 , 0) , (0001 , 0010) , 0001) ∷
((0 , 0) , (0001 , 0010) , 0001) ∷
((1 , 0) , (0010 , 0011) , 0001) ∷
((0 , 0) , (0010 , 0011) , 0010) ∷
((1 , 0) , (0011 , 0101) , 0010) ∷
((0 , 0) , (0011 , 0101) , 0011) ∷
((1 , 0) , (0101 , 1000) , 0011) ∷
((0 , 1) , (0000 , 0001) , 0000) ∷ []
-}
