module code.examples.Identity where

-- revisit the imports
open import Agda.Builtin.String renaming (primStringAppend to _+++_)
open import Agda.Primitive
open import Data.Bool renaming (Bool to Bit; false to 0#; true to 1#)
open import Data.Fin using (Fin; zero; suc)
open import Data.List hiding (lookup ; [_] ; and)
open import Data.Maybe hiding (map)
open import Data.Nat using (ℕ; _∸_; _<_ ; _≤_ ; _+_) renaming (zero to Z; suc to S)
open import Data.Product hiding (map ; swap ; curry ; uncurry)
open import Data.Vec hiding (foldl ; foldr ; _++_ ; map ; _[_]=_ ; length) renaming (lookup to vecref)

open import code.common.Idioms hiding (_∘_)
open import code.common.Contexts
open import code.common.MachineWords
open import code.common.Stream renaming (hd to shd; tl to stl ; map to mapStr ; zip to zipStr)

open import code.firrtl.Types
open import code.firrtl.Syntax
open import code.firrtl.TransCircuit

open import code.devicecalculus.MealyMachine
open import code.devicecalculus.Semantics


-- Some abbreviations

W1 : Ty
W1 = ↑ UInt 1

Δid : ModTy
Δid = Mod (W1 ⊗ Unit) , Unit , (W1 ⊗ Unit)  ⋊ []

oid : [] ⊢ₓ UInt 1 // Δid
oid = inp Zᵍ

idmod : [] ⊢ᵐ Δid
idmod = mod ◇ (node none (⊗case oid unit)) (node none unit) (node none unit)

id : [] ⊢ᶜ Δid
id = mod (mod ◇
             (node none (⊗case oid unit))
             (node none unit)
             (node none unit))

{-
circuit identity :
  module identity :
    input in    : UInt<1>
    input clk   : Clock
    output out  : UInt<1>

    out <= in
-}

id' : Mealy (Vec Bit 1 × ⊤) ⊤ (Vec Bit 1 × ⊤)
id' = ⟦ trC Δid [] id ⟧ []

insid : Stream (Vec Bit 1 × ⊤)
insid = cycle (0# ∷ [] , tt) ((1# ∷ [] , tt) ∷ (1# ∷ [] , tt) ∷ (1# ∷ [] , tt) ∷ [])

-- these are the interactive inputs.
inputs = map (λ (b , _) → ppBV b) (takeStr 20 insid)

-- hideous
pp : ((Vec Bit 1 × ⊤ {lzero}) × ⊤ {lzero} × Vec Bit 1 × ⊤ {lzero}) → String
pp ((ib , _) , s , (ob , _)) = (("(" +++ ((ppBV ib) +++ (" , " +++ ("tt" +++ " , ")))) +++ (ppBV ob)) +++ ")"

runid = map pp (takeStr 20 (transcript id' insid))


{-
"(0 , tt , 0)" ∷
"(0 , tt , 0)" ∷
"(1 , tt , 1)" ∷
"(1 , tt , 1)" ∷
"(1 , tt , 1)" ∷
"(0 , tt , 0)" ∷
"(1 , tt , 1)" ∷
"(1 , tt , 1)" ∷
"(1 , tt , 1)" ∷
"(0 , tt , 0)" ∷
"(1 , tt , 1)" ∷
"(1 , tt , 1)" ∷
"(1 , tt , 1)" ∷
"(0 , tt , 0)" ∷
"(1 , tt , 1)" ∷
"(1 , tt , 1)" ∷
"(1 , tt , 1)" ∷
"(0 , tt , 0)" ∷ 
"(1 , tt , 1)" ∷ 
"(1 , tt , 1)" ∷ []

-- un-prettyprinted
((0# ∷ [] , tt) , tt , 0# ∷ [] , tt) ∷
((0# ∷ [] , tt) , tt , 0# ∷ [] , tt) ∷
((1# ∷ [] , tt) , tt , 1# ∷ [] , tt) ∷
((1# ∷ [] , tt) , tt , 1# ∷ [] , tt) ∷
((1# ∷ [] , tt) , tt , 1# ∷ [] , tt) ∷
((0# ∷ [] , tt) , tt , 0# ∷ [] , tt) ∷
((1# ∷ [] , tt) , tt , 1# ∷ [] , tt) ∷
((1# ∷ [] , tt) , tt , 1# ∷ [] , tt) ∷
((1# ∷ [] , tt) , tt , 1# ∷ [] , tt) ∷
((0# ∷ [] , tt) , tt , 0# ∷ [] , tt) ∷
((1# ∷ [] , tt) , tt , 1# ∷ [] , tt) ∷
((1# ∷ [] , tt) , tt , 1# ∷ [] , tt) ∷
((1# ∷ [] , tt) , tt , 1# ∷ [] , tt) ∷
((0# ∷ [] , tt) , tt , 0# ∷ [] , tt) ∷
((1# ∷ [] , tt) , tt , 1# ∷ [] , tt) ∷
((1# ∷ [] , tt) , tt , 1# ∷ [] , tt) ∷
((1# ∷ [] , tt) , tt , 1# ∷ [] , tt) ∷
((0# ∷ [] , tt) , tt , 0# ∷ [] , tt) ∷
((1# ∷ [] , tt) , tt , 1# ∷ [] , tt) ∷
((1# ∷ [] , tt) , tt , 1# ∷ [] , tt) ∷ []
-}
