module code.examples.SanityChecks where

open import Agda.Builtin.String renaming (primStringAppend to _+++_) -- hiding (_≈_ ; _++_)
open import Relation.Binary.PropositionalEquality
open import Data.Bool renaming (Bool to Bit; false to 0#; true to 1#)
open import Data.Nat -- using (ℕ; _∸_; _<_ ; _≤_ ; _+_) renaming (zero to Z; suc to S)
open import Data.Product


open import code.common.Idioms 
open import code.common.MachineWords
open import code.common.Stream renaming (hd to shd; tl to stl ; map to mapStr ; zip to zipStr)
open import code.devicecalculus.MealyMachine

--
-- Small sanity check.
--

fibgen : Bit → ℕ × ℕ → (ℕ × (ℕ × ℕ))
fibgen 0# (n , m) = (n , (n , m))
fibgen 1# (n , m)  = (m , (m , n + m))

fib : Mealy Bit (ℕ × ℕ) ℕ 
fib = unfoldM fibgen 0# (0 , 1) 0

genfib : (Bit × Bit) → ℕ × ℕ → (ℕ × (ℕ × ℕ))
genfib (_ , 1#) (_ , _) = (0 , (0 , 1))
genfib (0# , _) (n , m) = (n , (n , m))
genfib (1# , _) (n , m) = (m , (m , n + m))

-- fib with reset
rfib : Mealy (Bit × Bit) (ℕ × ℕ) ℕ 
rfib = unfoldM genfib (0# , 0#) (0 , 0) 0

W8 = BitVector 8


genfib8 : (Bit × Bit) → W8 × W8 → (W8 × (W8 × W8))
genfib8 (_ , 1#) (_ , _) = (val0 , (val0 , val1))
genfib8 (0# , _) (n , m) = (n , (n , m))
genfib8 (1# , _) (n , m) = (m , (m , n ⊞ m))


W4 = BitVector 4

genfib4 : (Bit × Bit) → W4 × W4 → (W4 × (W4 × W4))
genfib4 (_ , 1#) (_ , _) = (v0 , (v0 , v1))
genfib4 (0# , _) (n , m) = (n , (n , m))
genfib4 (1# , _) (n , m) = (m , (m , n ⊞ m))

fib4 : Mealy (Bit × Bit) (W4 × W4) W4
fib4 = unfoldM genfib4 (0# , 0#) (v0 , v0) v0

rfib8 : Mealy (Bit × Bit) (W8 × W8) W8
rfib8 = unfoldM genfib8 (0# , 0#) (val0 , val0) val0

ins : Stream Bit
shd ins = 1#
stl ins = ins

alternating : Stream Bit
alternating = unfoldStr (λ s → s) (λ s → not s) 1#

runfib : Stream (Bit × (ℕ × ℕ) × ℕ)
runfib = transcript fib ins

runfib' : Stream (Bit × (ℕ × ℕ) × ℕ)
runfib' = transcript fib (0# ◁ (0# ◁ ins))

runfib2 : Stream (Bit × (ℕ × ℕ) × ℕ)
runfib2 = transcript fib alternating

open import Data.List

r1  = takeStr 1 runfib
r10 = takeStr 10 runfib

r1'  = takeStr 1 runfib'
r10' = takeStr 10 runfib'

t2 = takeStr 10 runfib2

seed1 : List (Bit × Bit)
seed1 =             (1# , 0#) ∷
        (0# , 0#) ∷ (1# , 0#) ∷ 
        (0# , 0#) ∷ (1# , 0#) ∷ 
        (0# , 0#) ∷ (1# , 0#) ∷ 
        (0# , 1#) ∷ (1# , 0#) ∷ 
        (0# , 0#) ∷ (1# , 0#) ∷ 
        (0# , 0#) ∷ (1# , 0#) ∷ 
        (0# , 0#) ∷ (1# , 0#) ∷ 
        (0# , 0#) ∷ (1# , 0#) ∷ 
        (0# , 0#) ∷ (1# , 1#) ∷ []


ppBit : Bit → String
ppBit b = primShowChar (bit2char b)

-- you would not believe how hard this was to write.
pp : ((Bit × Bit) × (W8 × W8) × W8) → String
pp ((b1 , b2) , (w1 , w2) , w3 ) = "(" +++ (ppBit b1 +++ (" , " +++ (ppBit b2 +++ (") , (" +++
                                    (ppBV w1 +++ (" , " +++ (ppBV w2 +++ (") , " +++
                                    (ppBV w3 +++ ")")))))))))

pp4 : ((Bit × Bit) × (W4 × W4) × W4) → String
pp4 ((b1 , b2) , (w1 , w2) , w3 ) = "(" +++ (ppBit b1 +++ (" , " +++ (ppBit b2 +++ (") , (" +++
                                    (ppBV w1 +++ (" , " +++ (ppBV w2 +++ (") , " +++
                                    (ppBV w3 +++ ")")))))))))

-------------------------------------------------------------
-------------------------------------------------------------
-- Below examples from the paper, run0 and run0'; seed0 defined
-- in common/Stream
-------------------------------------------------------------
-------------------------------------------------------------

run0 = runM 12 rfib (cycle (0# , 1#) seed0)

{- run0 output
((0# , 0#) , (0 , 0) , 0) ∷
((0# , 1#) , (0 , 1) , 0) ∷
((1# , 0#) , (1 , 1) , 1) ∷
((0# , 0#) , (1 , 1) , 1) ∷
((1# , 0#) , (1 , 2) , 1) ∷
((0# , 0#) , (1 , 2) , 1) ∷
((1# , 0#) , (2 , 3) , 2) ∷
((0# , 0#) , (2 , 3) , 2) ∷
((1# , 0#) , (3 , 5) , 3) ∷
((0# , 0#) , (3 , 5) , 3) ∷
((1# , 0#) , (5 , 8) , 5) ∷ 
((0# , 1#) , (0 , 1) , 0) ∷ []
-}

runningex4 : List String
runningex4 = Data.List.map pp4 (runM 12 fib4 (cycle (0# , 1#) seed0))

{-
"('0' , '0') , (0000 , 0000) , 0000)" ∷
"('0' , '1') , (0000 , 0001) , 0000)" ∷
"('1' , '0') , (0001 , 0001) , 0001)" ∷
"('0' , '0') , (0001 , 0001) , 0001)" ∷
"('1' , '0') , (0001 , 0010) , 0001)" ∷
"('0' , '0') , (0001 , 0010) , 0001)" ∷
"('1' , '0') , (0010 , 0011) , 0010)" ∷
"('0' , '0') , (0010 , 0011) , 0010)" ∷
"('1' , '0') , (0011 , 0101) , 0011)" ∷
"('0' , '0') , (0011 , 0101) , 0011)" ∷
"('1' , '0') , (0101 , 1000) , 0101)" ∷
"('0' , '1') , (0000 , 0001) , 0000)" ∷ []
-}

{- pretty printed by hand

((0 , 0) , (0000 , 0000) , 0000) ∷
((0 , 1) , (0000 , 0001) , 0000) ∷
((1 , 0) , (0001 , 0001) , 0001) ∷
((0 , 0) , (0001 , 0001) , 0001) ∷
((1 , 0) , (0001 , 0010) , 0001) ∷
((0 , 0) , (0001 , 0010) , 0001) ∷
((1 , 0) , (0010 , 0011) , 0010) ∷
((0 , 0) , (0010 , 0011) , 0010) ∷
((1 , 0) , (0011 , 0101) , 0011) ∷
((0 , 0) , (0011 , 0101) , 0011) ∷
((1 , 0) , (0101 , 1000) , 0101) ∷
((0 , 1) , (0000 , 0001) , 0000) ∷ []
-}

run0' : List String
run0' = Data.List.map pp (takeStr 12 (transcript rfib8 (cycle (0# , 1#) seed0)))

{- run0' output
"('0' , '0') , (00000000 , 00000000) , 00000000)" ∷
"('0' , '1') , (00000000 , 00000001) , 00000000)" ∷
"('1' , '0') , (00000001 , 00000001) , 00000001)" ∷
"('0' , '0') , (00000001 , 00000001) , 00000001)" ∷
"('1' , '0') , (00000001 , 00000010) , 00000001)" ∷
"('0' , '0') , (00000001 , 00000010) , 00000001)" ∷
"('1' , '0') , (00000010 , 00000011) , 00000010)" ∷
"('0' , '0') , (00000010 , 00000011) , 00000010)" ∷
"('1' , '0') , (00000011 , 00000101) , 00000011)" ∷
"('0' , '0') , (00000011 , 00000101) , 00000011)" ∷
"('1' , '0') , (00000101 , 00001000) , 00000101)" ∷
"('0' , '1') , (00000000 , 00000001) , 00000000)" ∷ []
-}

{-
run1 = takeStr 12 (transcript rfib (cycle (0# , 0#) seed1))
{- run1 output
((0# , 0#) , (0 , 1) , 0) ∷
((0# , 0#) , (0 , 1) , 0) ∷
((1# , 0#) , (1 , 1) , 1) ∷
((0# , 0#) , (1 , 1) , 1) ∷
((1# , 0#) , (1 , 2) , 1) ∷
((0# , 0#) , (1 , 2) , 1) ∷
((1# , 0#) , (2 , 3) , 2) ∷
((0# , 0#) , (2 , 3) , 2) ∷
((1# , 0#) , (3 , 5) , 3) ∷ 
((0# , 1#) , (0 , 1) , 0) ∷ []
-}

run1' = Data.List.map pp (takeStr 12 (transcript rfib8 (cycle (0# , 0#) seed1)))
{- run1' output
"('0' , '0') , (00000000 , 00000001) , 00000000)" ∷
"('0' , '0') , (00000000 , 00000001) , 00000000)" ∷
"('1' , '0') , (00000001 , 00000001) , 00000001)" ∷
"('0' , '0') , (00000001 , 00000001) , 00000001)" ∷
"('1' , '0') , (00000001 , 00000010) , 00000001)" ∷
"('0' , '0') , (00000001 , 00000010) , 00000001)" ∷
"('1' , '0') , (00000010 , 00000011) , 00000010)" ∷
"('0' , '0') , (00000010 , 00000011) , 00000010)" ∷
"('1' , '0') , (00000011 , 00000101) , 00000011)" ∷
"('0' , '0') , (00000011 , 00000101) , 00000011)" ∷ []
-}
-- Data.List.map pp run0'
-}
--
-- End of small sanity check.
--
