module code.examples.ShiftRegister where

-- revisit the imports
open import Agda.Builtin.String renaming (primStringAppend to _+++_)
open import Agda.Primitive
open import Data.Bool renaming (Bool to Bit; false to 0#; true to 1#)
open import Data.Fin using (Fin; zero; suc)
open import Data.List hiding (lookup ; [_] ; and)
open import Data.Maybe hiding (map)
open import Data.Nat using (ℕ; _∸_; _<_ ; _≤_ ; _+_) renaming (zero to Z; suc to S)
open import Data.Product hiding (map ; swap ; curry ; uncurry)
open import Data.Vec hiding (foldl ; foldr ; _++_ ; map ; _[_]=_ ; length) renaming (lookup to vecref ; reverse to rev)

open import code.common.Idioms hiding (_∘_)
open import code.common.Contexts
open import code.common.MachineWords
open import code.common.Stream renaming (hd to shd; tl to stl ; map to mapStr ; zip to zipStr)

open import code.firrtl.Types
open import code.firrtl.Syntax
open import code.firrtl.TransCircuit

open import code.devicecalculus.Types
open import code.devicecalculus.Syntax
open import code.devicecalculus.MealyMachine
open import code.devicecalculus.Semantics


-- Some abbreviations

W1 : Ty
W1 = ↑ UInt 1


-- circuit ShiftRegister : @[:@2.0]
--   module ShiftRegister : @[:@2.0]
--     input clock : Clock @[:@4.4]
--     input reset : UInt<1> @[:@5.4]
--     input io_in : UInt<1> @[:@6.4]
--     output io_out : UInt<1> @[:@6.4]
--
--     reg r0 : UInt<1>, clock with :
--       reset => (UInt<1>("h0"), r0) @[ShiftRegister.scala 11:19:@8.4]
--     reg r1 : UInt<1>, clock with :
--       reset => (UInt<1>("h0"), r1) @[ShiftRegister.scala 12:19:@10.4]
--     reg r2 : UInt<1>, clock with :
--       reset => (UInt<1>("h0"), r2) @[ShiftRegister.scala 13:19:@12.4]
--     reg r3 : UInt<1>, clock with :
--       reset => (UInt<1>("h0"), r3) @[ShiftRegister.scala 14:19:@14.4]
--     io_out <= r3 @[ShiftRegister.scala 15:10:@16.4]
--     r0 <= io_in @[ShiftRegister.scala 11:19:@9.4]
--     r1 <= r0 @[ShiftRegister.scala 12:19:@11.4]
--     r2 <= r1 @[ShiftRegister.scala 13:19:@13.4]
--     r3 <= r2 @[ShiftRegister.scala 14:19:@15.4]
--     skip

Δ : ModTy
Δ = Mod
        (W1 ⊗  -- reset
         W1 ⊗  -- io_in 
         Unit)
      ,
        (W1 ⊗  -- r0
         W1 ⊗  -- r1
         W1 ⊗  -- r2
         W1 ⊗  -- r3
         Unit)
      ,  (W1 ⊗ Unit) -- io_out
         ⋊ []

-- Next value of io_out is r3
io_out : [] ⊢ₓ UInt 1 // Δ
io_out = reg (Sᵍ (Sᵍ (Sᵍ Zᵍ)))

-- Next value of r0
ioin : [] ⊢ₓ UInt 1 // Δ
ioin = inp (Sᵍ Zᵍ)

-- Next value of r1
r0 : [] ⊢ₓ UInt 1 // Δ
r0 = reg Zᵍ

-- Next value of r2
r1 : [] ⊢ₓ UInt 1 // Δ
r1 = reg (Sᵍ Zᵍ)

-- Next value of r3
r2 : [] ⊢ₓ UInt 1 // Δ
r2 = reg (Sᵍ (Sᵍ Zᵍ))

shiftreg_mod : [] ⊢ᵐ Δ
shiftreg_mod = mod ◇
  {- outs -} (node none (⊗case io_out unit)) 
  {- regs -} (node none (⊗case ioin (⊗case r0 (⊗case r1 (⊗case r2 unit)))))
             (node none unit)

shiftreg : [] ⊢ᶜ Δ
shiftreg = mod shiftreg_mod

shiftregTr : [] ⊩ᵈ Dev (is Δ) (rs Δ) (os Δ)
shiftregTr = trC Δ [] shiftreg

shiftreg' : Mealy (Vec Bit 1 × Vec Bit 1 × ⊤)
                  (Vec Bit 1 × Vec Bit 1 × Vec Bit 1 × Vec Bit 1 × ⊤)
                  (Vec Bit 1 × ⊤)
shiftreg' = ⟦ shiftregTr ⟧ []

-- |
-- | Testing
-- |

zero1 : Vec Bit 1
zero1 = 0# ∷ []

one1 : Vec Bit 1
one1  = 1# ∷ []

-- | 
-- | These match the tests from ShiftRegister.{lo.fir,inputs,outputs}
-- |
testinputs : Stream (Vec Bit 1 × Vec Bit 1 × ⊤ {lzero})
testinputs = cycle testhd testtl
   where
     testhd : (Vec Bit 1 × Vec Bit 1 × ⊤ {lzero})
     testhd = (zero1 , zero1 , tt)  
     testtl : List (Vec Bit 1 × Vec Bit 1 × ⊤ {lzero})
     testtl = (zero1 , one1 , tt) ∷ 
              (zero1 , zero1 , tt) ∷ 
              (zero1 , one1 , tt) ∷ 
              (zero1 , one1 , tt) ∷ 
              (zero1 , zero1 , tt) ∷ 
              (zero1 , zero1 , tt) ∷ 
              (zero1 , zero1 , tt) ∷ 
              (zero1 , zero1 , tt) ∷ 
              (zero1 , one1 , tt) ∷ []
{-
Inputs distilled from simulator output
(0 , 0) :
(0 , 1) :
(0 , 0) :
(0 , 1) :
(0 , 1) :
(0 , 0) :
(0 , 0) :
(0 , 0) :
(0 , 0) :
(0 , 1) : []
-}


pair : String → String → String
pair i s = "(" +++ (i +++ (" , " +++ (s +++ ")")))

triple : String → String → String → String
triple i s o = "(" +++ (i +++ (" , " +++ (s +++ (" , " +++ (o +++ ")")))))

quad : String → String → String → String → String
quad i s o x = "(" +++ (i +++ (" , " +++
                       (s +++ (" , " +++
                       (o +++ (" , " +++
                       (x +++ ")")
                          ))
                            ))
                              ))
-- hideous
pp : (Vec Bit 1 × Vec Bit 1 × ⊤ {lzero}) × (Vec Bit 1 × Vec Bit 1 × Vec Bit 1 × Vec Bit 1 × ⊤ {lzero}) × (Vec Bit 1 × ⊤ {lzero})
         → String
pp ( ( reset , ioin , _) , ( r0 , r1 , r2 , r3 , _) , (ioout , _) )
        = triple
             (pair (ppBV reset) (ppBV ioin))
             (quad (ppBV r0) (ppBV r1) (ppBV r2) (ppBV r3))
             (ppBV ioout)

-- | Matches the simulator exactly.
runshiftreg = Data.List.tail (map pp (takeStr 11 (transcript shiftreg' testinputs)))

{-
just
("((0 , 0) , (0 , 0 , 0 , 0) , 0)" ∷
 "((0 , 1) , (1 , 0 , 0 , 0) , 0)" ∷
 "((0 , 0) , (0 , 1 , 0 , 0) , 0)" ∷
 "((0 , 1) , (1 , 0 , 1 , 0) , 0)" ∷
 "((0 , 1) , (1 , 1 , 0 , 1) , 0)" ∷ -- *
 "((0 , 0) , (0 , 1 , 1 , 0) , 1)" ∷ -- *
 "((0 , 0) , (0 , 0 , 1 , 1) , 0)" ∷ -- *
 "((0 , 0) , (0 , 0 , 0 , 1) , 1)" ∷ -- *
 "((0 , 0) , (0 , 0 , 0 , 0) , 1)" ∷
 "((0 , 1) , (1 , 0 , 0 , 0) , 0)" ∷ [])
-}

{-

CircuitState 5 (FRESH)
Inputs: clock= 0, io_in= 0, reset= 0
Outputs: io_out= 0
Registers      : r0= 0, r1= 0, r2= 0, r3= 0

CircuitState 6 (FRESH)
Inputs: clock= 0, io_in= 1, reset= 0
Outputs: io_out= 0
Registers      : r0= 1, r1= 0, r2= 0, r3= 0

CircuitState 7 (FRESH)
Inputs: clock= 0, io_in= 0, reset= 0
Outputs: io_out= 0
Registers      : r0= 0, r1= 1, r2= 0, r3= 0

CircuitState 8 (FRESH)
Inputs: clock= 0, io_in= 1, reset= 0
Outputs: io_out= 0
Registers      : r0= 1, r1= 0, r2= 1, r3= 0

CircuitState 9 (FRESH)
Inputs: clock= 0, io_in= 1, reset= 0
Outputs: io_out= 1
Registers      : r0= 1, r1= 1, r2= 0, r3= 1

CircuitState 10 (FRESH)
Inputs: clock= 0, io_in= 0, reset= 0
Outputs: io_out= 0
Registers      : r0= 0, r1= 1, r2= 1, r3= 0

CircuitState 11 (FRESH)
Inputs: clock= 0, io_in= 0, reset= 0
Outputs: io_out= 1
Registers      : r0= 0, r1= 0, r2= 1, r3= 1

CircuitState 12 (FRESH)
Inputs: clock= 0, io_in= 0, reset= 0
Outputs: io_out= 1
Registers      : r0= 0, r1= 0, r2= 0, r3= 1

CircuitState 13 (FRESH)
Inputs: clock= 0, io_in= 0, reset= 0
Outputs: io_out= 0
Registers      : r0= 0, r1= 0, r2= 0, r3= 0

CircuitState 14 (FRESH)
Inputs: clock= 0, io_in= 1, reset= 0
Outputs: io_out= 0
Registers      : r0= 1, r1= 0, r2= 0, r3= 0

-}
