module code.firrtl.Types where

open import Data.List
open import Data.Maybe
open import Data.Nat using (ℕ; zero; suc; _∸_; _<_ ; _≤_ ; _+_)
open import Data.Product hiding (map)

open import code.common.Contexts

--------------------------------------------------
--------------------------------------------------
-- FIRRTL Type syntax
--------------------------------------------------
--------------------------------------------------

--
-- Only ground types with fixed widths
--
data GrTy : Set where
   UInt  : ℕ → GrTy
   SInt  : ℕ → GrTy
   Clock : GrTy

infixr 4 _⊗_

--
--
--

data Ty : Set where
  ↑_   : GrTy → Ty
  _⊗_  : Ty → Ty → Ty
  Unit : Ty

--
--
--

infix 4 Mod_,_,_⋊_

data ModTy : Set where
  Mod_,_,_⋊_ : Ty → Ty → Ty → Cxt ModTy → ModTy

--
-- Indices for a GrTy in a Ty context.
--

infix 3 _∈ᵍ_

data _∈ᵍ_ (t : GrTy) : Ty → Set where
   Zᵍ : ∀ {τ} → t ∈ᵍ ((↑ t) ⊗ τ)
   Sᵍ : ∀ {τ τ'} → t ∈ᵍ τ' → t ∈ᵍ (τ ⊗ τ')

--
--
--

infix 3 _∈ᵗ_

data _∈ᵗ_ (t : Ty) : Ty → Set where
   Zᵗ : ∀ {τ} → t ∈ᵗ (t ⊗ τ)
   Sᵗ : ∀ {τ τ'} → t ∈ᵗ τ' → t ∈ᵗ (τ ⊗ τ')

--
-- Products on Ty
--

Πty : Ty → Cxt Ty → Ty
Πty h []       = h
Πty h (t ∷ []) = h ⊗ t
Πty h (t ∷ ts) = h ⊗ (Πty t ts)

